@@ -30,6 +30,8 @@
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/bits.h>
 #include <osmocom/core/utils.h>
+#include <osmocom/core/rate_ctr.h>
+#include <osmocom/core/stats.h>
 
 #include <osmocom/gsm/protocol/gsm_08_58.h>
 #include <osmocom/gsm/a5.h>
@@ -40,6 +42,7 @@
 #include <osmo-bts/l1sap.h>
 #include <osmo-bts/scheduler.h>
 #include <osmo-bts/scheduler_backend.h>
+#include <osmo-bts/bts.h>
 
 extern void *tall_bts_ctx;
 
@@ -49,8 +52,9 @@ static int rts_tchf_fn(struct l1sched_trx *l1t, uint8_t tn, uint32_t fn,
 	enum trx_chan_type chan);
 static int rts_tchh_fn(struct l1sched_trx *l1t, uint8_t tn, uint32_t fn,
 	enum trx_chan_type chan);
+
 /*! \brief Dummy Burst (TS 05.02 Chapter 5.2.6) */
-static const ubit_t dummy_burst[GSM_BURST_LEN] = {
+const ubit_t _sched_dummy_burst[GSM_BURST_LEN] = {
 	0,0,0,
 	1,1,1,1,1,0,1,1,0,1,1,1,0,1,1,0,0,0,0,0,1,0,1,0,0,1,0,0,1,1,1,0,
 	0,0,0,0,1,0,0,1,0,0,0,1,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,1,1,1,0,0,
@@ -60,15 +64,6 @@ static const ubit_t dummy_burst[GSM_BURST_LEN] = {
 	0,0,0,
 };
 
-/*! \brief FCCH Burst (TS 05.02 Chapter 5.2.4) */
-const ubit_t _sched_fcch_burst[GSM_BURST_LEN] = {
-	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
-	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
-	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
-	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
-	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
-};
-
 /*! \brief Training Sequences (TS 05.02 Chapter 5.2.3) */
 const ubit_t _sched_tsc[8][26] = {
 	{ 0,0,1,0,0,1,0,1,1,1,0,0,0,0,1,0,0,0,1,0,0,1,0,1,1,1, },
@@ -365,7 +360,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 		.ul_fn = rx_data_fn,
-		.nope_fn = rx_data_fn,
 	},
 	[TRXC_SACCHTH_0] = {
 		.name = "SACCH/TH(0)", /* 3GPP TS 05.02, section 3.3.4.1 */
@@ -377,7 +371,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 		.ul_fn = rx_data_fn,
-		.nope_fn = rx_data_fn,
 	},
 	[TRXC_SACCHTH_1] = {
 		.name = "SACCH/TH(1)", /* 3GPP TS 05.02, section 3.3.4.1 */
@@ -389,7 +382,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 		.ul_fn = rx_data_fn,
-		.nope_fn = rx_data_fn,
 	},
 	[TRXC_SACCH4_0] = {
 		.name = "SACCH/4(0)", /* 3GPP TS 05.02, section 3.3.4.1 */
@@ -401,7 +393,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 		.ul_fn = rx_data_fn,
-		.nope_fn = rx_data_fn,
 	},
 	[TRXC_SACCH4_1] = {
 		.name = "SACCH/4(1)", /* 3GPP TS 05.02, section 3.3.4.1 */
@@ -413,7 +404,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 		.ul_fn = rx_data_fn,
-		.nope_fn = rx_data_fn,
 	},
 	[TRXC_SACCH4_2] = {
 		.name = "SACCH/4(2)", /* 3GPP TS 05.02, section 3.3.4.1 */
@@ -425,7 +415,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 		.ul_fn = rx_data_fn,
-		.nope_fn = rx_data_fn,
 	},
 	[TRXC_SACCH4_3] = {
 		.name = "SACCH/4(3)", /* 3GPP TS 05.02, section 3.3.4.1 */
@@ -437,7 +426,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 		.ul_fn = rx_data_fn,
-		.nope_fn = rx_data_fn,
 	},
 	[TRXC_SACCH8_0] = {
 		.name = "SACCH/8(0)", /* 3GPP TS 05.02, section 3.3.4.1 */
@@ -449,7 +437,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 		.ul_fn = rx_data_fn,
-		.nope_fn = rx_data_fn,
 	},
 	[TRXC_SACCH8_1] = {
 		.name = "SACCH/8(1)", /* 3GPP TS 05.02, section 3.3.4.1 */
@@ -461,7 +448,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 		.ul_fn = rx_data_fn,
-		.nope_fn = rx_data_fn,
 	},
 	[TRXC_SACCH8_2] = {
 		.name = "SACCH/8(2)", /* 3GPP TS 05.02, section 3.3.4.1 */
@@ -473,7 +459,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 		.ul_fn = rx_data_fn,
-		.nope_fn = rx_data_fn,
 	},
 	[TRXC_SACCH8_3] = {
 		.name = "SACCH/8(3)", /* 3GPP TS 05.02, section 3.3.4.1 */
@@ -485,7 +470,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 		.ul_fn = rx_data_fn,
-		.nope_fn = rx_data_fn,
 	},
 	[TRXC_SACCH8_4] = {
 		.name = "SACCH/8(4)", /* 3GPP TS 05.02, section 3.3.4.1 */
@@ -497,7 +481,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 		.ul_fn = rx_data_fn,
-		.nope_fn = rx_data_fn,
 	},
 	[TRXC_SACCH8_5] = {
 		.name = "SACCH/8(5)", /* 3GPP TS 05.02, section 3.3.4.1 */
@@ -509,7 +492,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 		.ul_fn = rx_data_fn,
-		.nope_fn = rx_data_fn,
 	},
 	[TRXC_SACCH8_6] = {
 		.name = "SACCH/8(6)", /* 3GPP TS 05.02, section 3.3.4.1 */
@@ -521,7 +503,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 		.ul_fn = rx_data_fn,
-		.nope_fn = rx_data_fn,
 	},
 	[TRXC_SACCH8_7] = {
 		.name = "SACCH/8(7)", /* 3GPP TS 05.02, section 3.3.4.1 */
@@ -533,7 +514,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_data_fn,
 		.ul_fn = rx_data_fn,
-		.nope_fn = rx_data_fn,
 	},
 	[TRXC_PDTCH] = {
 		.name = "PDTCH", /* 3GPP TS 05.02, sections 3.2.4, 3.3.2.4 */
@@ -544,7 +524,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		 * 05.03, chapter 5), regular interleaving as specified for xCCH.
 		 * NOTE: the burst buffer is three times bigger because the
 		 * payload of EDGE bursts is three times longer. */
-		.flags = TRX_CHAN_FLAG_PDCH,
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_pdtch_fn,
 		.ul_fn = rx_pdtch_fn,
@@ -559,7 +538,6 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 		 * transfer mode. On Downlink, the network sends timing advance
 		 * updates for several mobile stations. The coding scheme used
 		 * for PTCCH/D messages is the same as for PDTCH CS-1. */
-		.flags = TRX_CHAN_FLAG_PDCH,
 		.rts_fn = rts_data_fn,
 		.dl_fn = tx_pdtch_fn,
 		.ul_fn = rx_rach_fn,
@@ -577,6 +555,23 @@ const struct trx_chan_desc trx_chan_desc[_TRX_CHAN_MAX] = {
 	},
 };
 
+enum {
+	L1SCHED_TS_CTR_DL_LATE,
+	L1SCHED_TS_CTR_DL_NOT_FOUND,
+};
+
+static const struct rate_ctr_desc l1sched_ts_ctr_desc[] = {
+	[L1SCHED_TS_CTR_DL_LATE] =	{"l1sched_ts:dl_late", "Downlink frames arrived too late to submit to lower layers"},
+	[L1SCHED_TS_CTR_DL_NOT_FOUND] =	{"l1sched_ts:dl_not_found", "Downlink frames not found while scheduling"},
+};
+static const struct rate_ctr_group_desc l1sched_ts_ctrg_desc = {
+	"l1sched_ts",
+	"L1 scheduler timeslot",
+	OSMO_STATS_CLASS_GLOBAL,
+	ARRAY_SIZE(l1sched_ts_ctr_desc),
+	l1sched_ts_ctr_desc
+};
+
 /*
  * init / exit
  */
@@ -597,11 +592,13 @@ int trx_sched_init(struct l1sched_trx *l1t, struct gsm_bts_trx *trx)
 		struct l1sched_ts *l1ts = l1sched_trx_get_ts(l1t, tn);
 
 		l1ts->mf_index = 0;
+		l1ts->ctrs = rate_ctr_group_alloc(trx, &l1sched_ts_ctrg_desc, (trx->nr + 1) * 10 + tn);
 		INIT_LLIST_HEAD(&l1ts->dl_prims);
+
 		for (i = 0; i < ARRAY_SIZE(l1ts->chan_state); i++) {
 			struct l1sched_chan_state *chan_state;
 			chan_state = &l1ts->chan_state[i];
-			chan_state->active = 0;
+			chan_state->active = false;
 		}
 	}
 
@@ -619,6 +616,8 @@ void trx_sched_exit(struct l1sched_trx *l1t)
 	for (tn = 0; tn < ARRAY_SIZE(l1t->ts); tn++) {
 		struct l1sched_ts *l1ts = l1sched_trx_get_ts(l1t, tn);
 		msgb_queue_flush(&l1ts->dl_prims);
+		rate_ctr_group_free(l1ts->ctrs);
+		l1ts->ctrs = NULL;
 		for (i = 0; i < _TRX_CHAN_MAX; i++) {
 			struct l1sched_chan_state *chan_state;
 			chan_state = &l1ts->chan_state[i];
@@ -638,80 +637,73 @@ void trx_sched_exit(struct l1sched_trx *l1t)
 	}
 }
 
-/* close all logical channels and reset timeslots */
-void trx_sched_reset(struct l1sched_trx *l1t)
-{
-	trx_sched_exit(l1t);
-	trx_sched_init(l1t, l1t->trx);
-}
-
 struct msgb *_sched_dequeue_prim(struct l1sched_trx *l1t, int8_t tn, uint32_t fn,
 				 enum trx_chan_type chan)
 {
 	struct msgb *msg, *msg2;
-	struct osmo_phsap_prim *l1sap;
-	uint32_t prim_fn;
+	uint32_t prim_fn, l1sap_fn;
 	uint8_t chan_nr, link_id;
 	struct l1sched_ts *l1ts = l1sched_trx_get_ts(l1t, tn);
 
 	/* get prim of current fn from queue */
 	llist_for_each_entry_safe(msg, msg2, &l1ts->dl_prims, list) {
-		l1sap = msgb_l1sap_prim(msg);
-		if (l1sap->oph.operation != PRIM_OP_REQUEST) {
-wrong_type:
-			LOGL1S(DL1P, LOGL_ERROR, l1t, tn, chan, fn, "Prim has wrong type.\n");
-free_msg:
-			/* unlink and free message */
-			llist_del(&msg->list);
-			msgb_free(msg);
-			return NULL;
-		}
+		struct osmo_phsap_prim *l1sap = msgb_l1sap_prim(msg);
 		switch (l1sap->oph.primitive) {
 		case PRIM_PH_DATA:
 			chan_nr = l1sap->u.data.chan_nr;
 			link_id = l1sap->u.data.link_id;
-			prim_fn = ((l1sap->u.data.fn + GSM_HYPERFRAME - fn) % GSM_HYPERFRAME);
+			l1sap_fn = l1sap->u.data.fn;
 			break;
 		case PRIM_TCH:
 			chan_nr = l1sap->u.tch.chan_nr;
 			link_id = 0;
-			prim_fn = ((l1sap->u.tch.fn + GSM_HYPERFRAME - fn) % GSM_HYPERFRAME);
+			l1sap_fn = l1sap->u.tch.fn;
 			break;
 		default:
-			goto wrong_type;
+			LOGL1S(DL1P, LOGL_ERROR, l1t, tn, chan, fn, "Prim has wrong type.\n");
+			goto free_msg;
 		}
-		if (prim_fn > 100) {
+		prim_fn = GSM_TDMA_FN_SUB(l1sap_fn, fn);
+		if (prim_fn > 100) { /* l1sap_fn < fn */
 			LOGL1S(DL1P, LOGL_NOTICE, l1t, tn, chan, fn,
-			     "Prim %u is out of range (100), or channel %s with "
+			     "Prim %u is out of range (%u vs exp %u), or channel %s with "
 			     "type %s is already disabled. If this happens in "
 			     "conjunction with PCU, increase 'rts-advance' by 5.\n",
-			     prim_fn, get_lchan_by_chan_nr(l1t->trx, chan_nr)->name,
+			     prim_fn, l1sap_fn, fn,
+			     get_lchan_by_chan_nr(l1t->trx, chan_nr)->name,
 			     trx_chan_desc[chan].name);
+			rate_ctr_inc2(l1ts->ctrs, L1SCHED_TS_CTR_DL_LATE);
 			/* unlink and free message */
 			llist_del(&msg->list);
 			msgb_free(msg);
 			continue;
 		}
-		if (prim_fn > 0)
-			continue;
+		if (prim_fn > 0) /* l1sap_fn > fn */
+			break;
 
-		goto found_msg;
+		/* l1sap_fn == fn */
+		if ((chan_nr ^ (trx_chan_desc[chan].chan_nr | tn))
+		 || ((link_id & 0xc0) ^ trx_chan_desc[chan].link_id)) {
+			LOGL1S(DL1P, LOGL_ERROR, l1t, tn, chan, fn, "Prim has wrong chan_nr=0x%02x link_id=%02x, "
+				"expecting chan_nr=0x%02x link_id=%02x.\n", chan_nr, link_id,
+				trx_chan_desc[chan].chan_nr | tn, trx_chan_desc[chan].link_id);
+			goto free_msg;
+		}
+
+		/* unlink and return message */
+		llist_del(&msg->list);
+		return msg;
 	}
 
+	/* Queue was traversed with no candidate, no prim is available for current FN: */
+	rate_ctr_inc2(l1ts->ctrs, L1SCHED_TS_CTR_DL_NOT_FOUND);
 	return NULL;
 
-found_msg:
-	if ((chan_nr ^ (trx_chan_desc[chan].chan_nr | tn))
-	 || ((link_id & 0xc0) ^ trx_chan_desc[chan].link_id)) {
-		LOGL1S(DL1P, LOGL_ERROR, l1t, tn, chan, fn, "Prim has wrong chan_nr=0x%02x link_id=%02x, "
-			"expecting chan_nr=0x%02x link_id=%02x.\n", chan_nr, link_id,
-			trx_chan_desc[chan].chan_nr | tn, trx_chan_desc[chan].link_id);
-		goto free_msg;
-	}
-
-	/* unlink and return message */
+free_msg:
+	/* unlink and free message */
 	llist_del(&msg->list);
-	return msg;
+	msgb_free(msg);
+	return NULL;
 }
 
 int _sched_compose_ph_data_ind(struct l1sched_trx *l1t, uint8_t tn, uint32_t fn,
@@ -753,7 +745,9 @@ int _sched_compose_ph_data_ind(struct l1sched_trx *l1t, uint8_t tn, uint32_t fn,
 }
 
 int _sched_compose_tch_ind(struct l1sched_trx *l1t, uint8_t tn, uint32_t fn,
-		    enum trx_chan_type chan, uint8_t *tch, uint8_t tch_len)
+			   enum trx_chan_type chan, uint8_t *tch, uint8_t tch_len,
+			   int16_t ta_offs_256bits, uint16_t ber10k, float rssi,
+			   uint8_t is_sub)
 {
 	struct msgb *msg;
 	struct osmo_phsap_prim *l1sap;
@@ -769,6 +763,11 @@ int _sched_compose_tch_ind(struct l1sched_trx *l1t, uint8_t tn, uint32_t fn,
 		PRIM_OP_INDICATION, msg);
 	l1sap->u.tch.chan_nr = chan_nr;
 	l1sap->u.tch.fn = fn;
+	l1sap->u.tch.rssi = (int8_t) (rssi);
+	l1sap->u.tch.ber10k = ber10k;
+	l1sap->u.tch.ta_offs_256bits = ta_offs_256bits;
+	l1sap->u.tch.is_sub = is_sub & 1;
+
 	msg->l2h = msgb_put(msg, tch_len);
 	if (tch_len)
 		memcpy(msg->l2h, tch, tch_len);
@@ -800,6 +799,7 @@ int trx_sched_ph_data_req(struct l1sched_trx *l1t, struct osmo_phsap_prim *l1sap
 		"PH-DATA.req: chan_nr=0x%02x link_id=0x%02x\n",
 		l1sap->u.data.chan_nr, l1sap->u.data.link_id);
 
+	OSMO_ASSERT(l1sap->oph.operation == PRIM_OP_REQUEST);
 	OSMO_ASSERT(l1sap->oph.msg);
 
 	/* ignore empty frame */
@@ -821,6 +821,7 @@ int trx_sched_tch_req(struct l1sched_trx *l1t, struct osmo_phsap_prim *l1sap)
 	LOGL1S(DL1P, LOGL_DEBUG, l1t, tn, -1, l1sap->u.tch.fn, "TCH.req: chan_nr=0x%02x\n",
 		l1sap->u.tch.chan_nr);
 
+	OSMO_ASSERT(l1sap->oph.operation == PRIM_OP_REQUEST);
 	OSMO_ASSERT(l1sap->oph.msg);
 
 	/* ignore empty frame */
@@ -835,7 +836,7 @@ int trx_sched_tch_req(struct l1sched_trx *l1t, struct osmo_phsap_prim *l1sap)
 }
 
 
-/* 
+/*
  * ready-to-send indication (to upper layer)
  */
 
@@ -857,6 +858,12 @@ static int rts_data_fn(struct l1sched_trx *l1t, uint8_t tn, uint32_t fn,
 		return -ENODEV;
 	}
 
+	/* For handover detection, there are cases where the SACCH should remain inactive until the first RACH
+	 * indicating the TA is received. */
+	if (L1SAP_IS_LINK_SACCH(link_id)
+	    && !l1t->ts[tn].chan_state[chan].lchan->want_dl_sacch_active)
+		return 0;
+
 	LOGL1S(DL1P, LOGL_DEBUG, l1t, tn, chan, fn,
 		"PH-RTS.ind: chan_nr=0x%02x link_id=0x%02x\n", chan_nr, link_id);
 
@@ -967,55 +974,93 @@ int trx_sched_set_pchan(struct l1sched_trx *l1t, uint8_t tn,
 	return 0;
 }
 
+/* Remove all matching (by chan_nr & link_id) primitives from the given queue */
+static void trx_sched_queue_filter(struct llist_head *q, uint8_t chan_nr, uint8_t link_id)
+{
+	struct msgb *msg, *_msg;
+
+	llist_for_each_entry_safe(msg, _msg, q, list) {
+		struct osmo_phsap_prim *l1sap = msgb_l1sap_prim(msg);
+		switch (l1sap->oph.primitive) {
+		case PRIM_PH_DATA:
+			if (l1sap->u.data.chan_nr != chan_nr)
+				continue;
+			if (l1sap->u.data.link_id != link_id)
+				continue;
+			break;
+		case PRIM_TCH:
+			if (l1sap->u.tch.chan_nr != chan_nr)
+				continue;
+			if (link_id != 0x00)
+				continue;
+			break;
+		default:
+			/* Shall not happen */
+			OSMO_ASSERT(0);
+		}
+
+		/* Unlink and free() */
+		llist_del(&msg->list);
+		talloc_free(msg);
+	}
+}
+
 /* setting all logical channels given attributes to active/inactive */
-int trx_sched_set_lchan(struct l1sched_trx *l1t, uint8_t chan_nr, uint8_t link_id,
-	int active)
+int trx_sched_set_lchan(struct l1sched_trx *l1t, uint8_t chan_nr, uint8_t link_id, bool active)
 {
 	uint8_t tn = L1SAP_CHAN2TS(chan_nr);
 	struct l1sched_ts *l1ts = l1sched_trx_get_ts(l1t, tn);
 	uint8_t ss = l1sap_chan2ss(chan_nr);
+	bool found = false;
 	int i;
-	int rc = -EINVAL;
 
 	/* look for all matching chan_nr/link_id */
 	for (i = 0; i < _TRX_CHAN_MAX; i++) {
-		struct l1sched_chan_state *chan_state;
-		chan_state = &l1ts->chan_state[i];
-		/* Skip if pchan type does not match pdch flag.
-		 * FIXME: Is it possible at all? Clarify if so. */
-		if ((trx_sched_multiframes[l1ts->mf_index].pchan == GSM_PCHAN_PDCH)
-		    && !(trx_chan_desc[i].flags & TRX_CHAN_FLAG_PDCH))
+		struct l1sched_chan_state *chan_state = &l1ts->chan_state[i];
+
+		if (trx_chan_desc[i].chan_nr != (chan_nr & RSL_CHAN_NR_MASK))
 			continue;
-		if (trx_chan_desc[i].chan_nr == (chan_nr & 0xf8)
-		 && trx_chan_desc[i].link_id == link_id) {
-			rc = 0;
-			if (chan_state->active == active)
-				continue;
-			LOGP(DL1C, LOGL_NOTICE, "%s %s on trx=%d ts=%d\n",
-				(active) ? "Activating" : "Deactivating",
-				trx_chan_desc[i].name, l1t->trx->nr, tn);
-			if (active)
-				memset(chan_state, 0, sizeof(*chan_state));
-			chan_state->active = active;
-			/* free burst memory, to cleanly start with burst 0 */
-			if (chan_state->dl_bursts) {
-				talloc_free(chan_state->dl_bursts);
-				chan_state->dl_bursts = NULL;
-			}
-			if (chan_state->ul_bursts) {
-				talloc_free(chan_state->ul_bursts);
-				chan_state->ul_bursts = NULL;
-			}
-			if (!active)
-				chan_state->ho_rach_detect = 0;
+		if (trx_chan_desc[i].link_id != link_id)
+			continue;
+		if (chan_state->active == active)
+			continue;
+		found = true;
+
+		LOGP(DL1C, LOGL_NOTICE, "%s %s on trx=%d ts=%d\n",
+			(active) ? "Activating" : "Deactivating",
+			trx_chan_desc[i].name, l1t->trx->nr, tn);
+		/* free burst memory, to cleanly start with burst 0 */
+		if (chan_state->dl_bursts) {
+			talloc_free(chan_state->dl_bursts);
+			chan_state->dl_bursts = NULL;
 		}
+		if (chan_state->ul_bursts) {
+			talloc_free(chan_state->ul_bursts);
+			chan_state->ul_bursts = NULL;
+		}
+
+		if (active) {
+			/* Clean up everything */
+			memset(chan_state, 0, sizeof(*chan_state));
+
+			/* Bind to generic 'struct gsm_lchan' */
+			chan_state->lchan = get_lchan_by_chan_nr(l1t->trx, chan_nr);
+			OSMO_ASSERT(chan_state->lchan != NULL);
+		} else {
+			chan_state->ho_rach_detect = 0;
+
+			/* Remove pending Tx prims belonging to this lchan */
+			trx_sched_queue_filter(&l1ts->dl_prims, chan_nr, link_id);
+		}
+
+		chan_state->active = active;
 	}
 
 	/* disable handover detection (on deactivation) */
 	if (!active)
 		_sched_act_rach_det(l1t, tn, ss, 0);
 
-	return rc;
+	return found ? 0 : -EINVAL;
 }
 
 /* setting all logical channels given attributes to active/inactive */
@@ -1099,10 +1144,7 @@ int trx_sched_set_cipher(struct l1sched_trx *l1t, uint8_t chan_nr, int downlink,
 
 	/* look for all matching chan_nr */
 	for (i = 0; i < _TRX_CHAN_MAX; i++) {
-		/* skip if pchan type */
-		if (trx_chan_desc[i].flags & TRX_CHAN_FLAG_PDCH)
-			continue;
-		if (trx_chan_desc[i].chan_nr == (chan_nr & 0xf8)) {
+		if (trx_chan_desc[i].chan_nr == (chan_nr & RSL_CHAN_NR_MASK)) {
 			chan_state = &l1ts->chan_state[i];
 			LOGP(DL1C, LOGL_NOTICE, "Set a5/%d %s for %s on trx=%d "
 				"ts=%d\n", algo,
@@ -1162,23 +1204,21 @@ int _sched_rts(struct l1sched_trx *l1t, uint8_t tn, uint32_t fn)
 }
 
 /* process downlink burst */
-const ubit_t *_sched_dl_burst(struct l1sched_trx *l1t, uint8_t tn,
-				uint32_t fn, uint16_t *nbits)
+void _sched_dl_burst(struct l1sched_trx *l1t, struct trx_dl_burst_req *br)
 {
-	struct l1sched_ts *l1ts = l1sched_trx_get_ts(l1t, tn);
+	struct l1sched_ts *l1ts = l1sched_trx_get_ts(l1t, br->tn);
 	struct l1sched_chan_state *l1cs;
 	const struct trx_sched_frame *frame;
 	uint8_t offset, period, bid;
 	trx_sched_dl_func *func;
 	enum trx_chan_type chan;
-	ubit_t *bits = NULL;
 
 	if (!l1ts->mf_index)
-		goto no_data;
+		return;
 
 	/* get frame from multiframe */
 	period = l1ts->mf_period;
-	offset = fn % period;
+	offset = br->fn % period;
 	frame = l1ts->mf_frames + offset;
 
 	chan = frame->dl_chan;
@@ -1188,48 +1228,30 @@ const ubit_t *_sched_dl_burst(struct l1sched_trx *l1t, uint8_t tn,
 	l1cs = &l1ts->chan_state[chan];
 
 	/* check if channel is active */
-	if (!TRX_CHAN_IS_ACTIVE(l1cs, chan)) {
-		if (nbits)
-			*nbits = GSM_BURST_LEN;
-		goto no_data;
-	}
+	if (!TRX_CHAN_IS_ACTIVE(l1cs, chan))
+		return;
 
 	/* get burst from function */
-	bits = func(l1t, tn, fn, chan, bid, nbits);
+	if (func(l1t, chan, bid, br) != 0)
+		return;
+
+	/* BS Power reduction (in dB) per logical channel */
+	if (l1cs->lchan != NULL)
+		br->att = l1cs->lchan->bs_power_ctrl.current;
 
 	/* encrypt */
-	if (bits && l1cs->dl_encr_algo) {
+	if (br->burst_len && l1cs->dl_encr_algo) {
 		ubit_t ks[114];
 		int i;
 
-		osmo_a5(l1cs->dl_encr_algo, l1cs->dl_encr_key, fn, ks, NULL);
+		osmo_a5(l1cs->dl_encr_algo, l1cs->dl_encr_key, br->fn, ks, NULL);
 		for (i = 0; i < 57; i++) {
-			bits[i + 3] ^= ks[i];
-			bits[i + 88] ^= ks[i + 57];
+			br->burst[i +  3] ^= ks[i];
+			br->burst[i + 88] ^= ks[i + 57];
 		}
 	}
-
-no_data:
-	/* in case of C0, we need a dummy burst to maintain RF power */
-	if (bits == NULL && l1t->trx == l1t->trx->bts->c0) {
-#if 0
-		if (chan != TRXC_IDLE) // hack
-			LOGP(DL1C, LOGL_DEBUG, "No burst data for %s fn=%u ts=%u "
-			     "burst=%d on C0, so filling with dummy burst\n",
-			     trx_chan_desc[chan].name, fn, tn, bid);
-#endif
-		bits = (ubit_t *) dummy_burst;
-	}
-
-	return bits;
 }
 
-#define TDMA_FN_SUM(a, b) \
-	((a + GSM_HYPERFRAME + b) % GSM_HYPERFRAME)
-
-#define TDMA_FN_SUB(a, b) \
-	((a + GSM_HYPERFRAME - b) % GSM_HYPERFRAME)
-
 static int trx_sched_calc_frame_loss(struct l1sched_trx *l1t,
 	struct l1sched_chan_state *l1cs, uint8_t tn, uint32_t fn)
 {
@@ -1267,7 +1289,7 @@ static int trx_sched_calc_frame_loss(struct l1sched_trx *l1t,
 	}
 
 	/* How many frames elapsed since the last one? */
-	elapsed_fs = TDMA_FN_SUB(fn, l1cs->last_tdma_fn);
+	elapsed_fs = GSM_TDMA_FN_SUB(fn, l1cs->last_tdma_fn);
 	if (elapsed_fs > l1ts->mf_period) { /* Too many! */
 		LOGL1S(DL1P, LOGL_ERROR, l1t, tn, frame_head->ul_chan, fn,
 			"Too many (>%u) contiguous TDMA frames=%u elapsed "
@@ -1286,7 +1308,7 @@ static int trx_sched_calc_frame_loss(struct l1sched_trx *l1t,
 	 * Start counting from the last_fn + 1.
 	 */
 	for (i = 1; i < elapsed_fs; i++) {
-		fn_i = TDMA_FN_SUM(l1cs->last_tdma_fn, i);
+		fn_i = GSM_TDMA_FN_SUM(l1cs->last_tdma_fn, i);
 		offset = fn_i % l1ts->mf_period;
 		frame = l1ts->mf_frames + offset;
 
@@ -1319,7 +1341,7 @@ static int trx_sched_calc_frame_loss(struct l1sched_trx *l1t,
 		};
 
 		for (i = 1; i < elapsed_fs; i++) {
-			fn_i = TDMA_FN_SUM(l1cs->last_tdma_fn, i);
+			fn_i = GSM_TDMA_FN_SUM(l1cs->last_tdma_fn, i);
 			offset = fn_i % l1ts->mf_period;
 			frame = l1ts->mf_frames + offset;
 			func = trx_chan_desc[frame->ul_chan].ul_fn;
@@ -1386,12 +1408,19 @@ int trx_sched_ul_burst(struct l1sched_trx *l1t, struct trx_ul_burst_ind *bi)
 	l1cs->last_tdma_fn = bi->fn;
 	l1cs->proc_tdma_fs++;
 
-	/* handle NOPE indications (if the handler is present) */
+	/* handle NOPE indications */
 	if (bi->flags & TRX_BI_F_NOPE_IND) {
-		func = trx_chan_desc[chan].nope_fn;
-		if (!func)
+		switch (chan) {
+		case TRXC_PDTCH:
+		case TRXC_PTCCH:
+		case TRXC_RACH:
+			/* For some logical channel types NOPE.ind is valueless. */
 			return 0;
-		return func(l1t, chan, bid, bi);
+		default:
+			/* NOTE: Uplink burst handler must check bi->burst_len before
+			 * accessing bi->burst to avoid uninitialized memory access. */
+			return func(l1t, chan, bid, bi);
+		}
 	}
 
 	/* decrypt */
